/****************************************************************************

	PROGRAM: WCookie.c

	PURPOSE: Windows Fortune Cookie

			 For those of you who have never seen the Unix/DOS cookie, it is
			 a rather simple program that selects a cute saying from a file
			 full of accumulated wisdom. This Windows-version uses the same
			 datafile as the Unix/DOS version so you can go on using your own
			 if you (like the undersigned) have added to it. One added bonus of this
			 windows-cookie is that it will go away automatically after a while.

             I Wrote this as an exercise in Windows-programming. I had
             been playing around with the idea of a Cookie for Windows for
             quite a while and when a friend of mine, who has artistic talent,
             promised to design a nice icon, I decided to dedicate a weekend to
             programming. Apart from the basic Windows-stuff, this application uses:
             - Timers to kill the application automatically after a while.
             - Child windows (the OK button)
             - Adding stuff to the system menu

	FUNCTIONS:

		WinMain() - calls initialization function, processes message loop
		InitApplication() - initializes window data and registers window
		InitInstance() - saves instance handle and creates main window
		MainWndProc() - processes messages
        About() - Show a box with program credits

	AUTHOR:	Anders Gustafsson (BIX:agustafsson). Cookie Icon by Nina Backlund.

	SUGGESTIONS:
		When I wrote this, I stood on the shoulders of others, who in their
		turn stood on others. By all means feel free to use this program or
		parts of it if you like. It has several weak areas that could be improved:
		1) Settings should be stored in win.ini instead of being hard-coded.
		2) The parsing of COOKIE.S could be improved
		3) Strings should be in the resources instead of in the code-segment
		4) Several thing that should be dependent on font-metrics are hard-coded.

		If you have comments to this program that you think could help me and
		other beginning Windows-programmers, please post them in prog.queries
		or e-mail me!

****************************************************************************/

#include <windows.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include <time.h>
#include "WCookie.h"		/* specific to this program 		*/

HANDLE hInst;				/* current instance kept global		*/
HANDLE hFile;               /* handle to cookie datafile		*/
HANDLE hBuffer;             /* Handle to memory holding cookie	*/
HANDLE hButton;             /* Handle to OK-button				*/
HANDLE idTimer;             /* Handle to timer that closes app	*/
HANDLE hMenu;				/* Handle to system menu			*/
FARPROC lpfnAbout;
HDC hDC;
PAINTSTRUCT ps;
OFSTRUCT OfStruct;
struct stat FileStatus;
char *String[80];			/* Array of pointers to each string in the cookie	*/
int	iStrings;               /* Number of strings in the cookie					*/
int iButtonWidth,iButtonHeight;
PSTR pBuffer;

/****************************************************************************

	FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)

	PURPOSE: calls initialization function, processes message loop

	COMMENTS:


****************************************************************************/

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;	/* message */

	if (!hPrevInstance)		/* Other instances of app running? */
		if (!InitApplication(hInstance))	/* Initialize shared things */
			return (FALSE);		/* Exits if unable to initialize */

	/* Perform initializations that apply to a specific instance */
	if (!InitInstance(hInstance, nCmdShow))
		return (FALSE);

	/* Acquire and dispatch messages until a WM_QUIT message is received. */
	while (GetMessage(&msg,		/* message structure */
		NULL,	/* handle of window receiving the message */
		NULL,	/* lowest message to examine	*/
		NULL))	/* highest message to examine	*/
	{
		TranslateMessage(&msg);		/* Translates virtual key codes		  */
		DispatchMessage(&msg);		/* Dispatches message to window		  */
	}
	return (msg.wParam);		/* Returns the value from PostQuitMessage */
}


/****************************************************************************

	FUNCTION: InitApplication(HANDLE)

	PURPOSE: Initializes window data and registers window class

	COMMENTS:

		This function is called at initialization time only if no other
		instances of the application are running.  This function performs
		initialization tasks that can be done once for any number of running
		instances.

		In this case, we initialize a window class by filling out a data
		structure of type WNDCLASS and calling the Windows RegisterClass()
		function. Since all instances of this application use the same window
		class, we only need to do this when the first instance is initialized.


****************************************************************************/

BOOL InitApplication(HANDLE hInstance)
{
	WNDCLASS  wc;

	/* Fill in window class structure with parameters that describe the
		main window. */

	wc.style = NULL;				/* Class style(s) */
	wc.lpfnWndProc = MainWndProc;	/* Function to retrieve messages for  */
									/* windows of this class */
	wc.cbClsExtra = 0;				/* No per-class extra data */
	wc.cbWndExtra = 0;				/* No per-window extra data */
	wc.hInstance = hInstance;		/* Application that owns the class */
	wc.hIcon = LoadIcon(hInstance, "COOKIE");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  NULL;	/* No menu. */
	wc.lpszClassName = "CookieClass"; /* Name used in call to CreateWindow. */

	/* Register the window class and return success/failure code. */
	return (RegisterClass(&wc));
}

/****************************************************************************

	FUNCTION:  InitInstance(HANDLE, int)

	PURPOSE:  Saves instance handle and creates main window

	COMMENTS:

		This function is called at initialization time for every instance of
		this application. Here we open the Cookie-file, locate a cookie and
		save it. The Cookie-algorithm is the same as in the Unix/DOS cookie:

		Open datafile. Create a random pointer into the datafile using rand().
		Start reading, looking for a Ctrl-A. When found note start of cookie.
		Continue reading until next Ctrl-A end of cookie. Cr/lfs are nulled
		and the start of each individual string is stored in an array.

		To simplify program logic we assume that no cookie is larger than
		80x25 characters, hence if we read 4k there should be at least two
		cookies.

****************************************************************************/

BOOL InitInstance(HANDLE hInstance, int nCmdShow)
{
	HWND	hWnd;		/* Main window handle */
    int		i;
    int		iBytes;
    unsigned long	lFilePos;
    unsigned int temp;
	/* Save the instance handle in static variable, which will be used in
	   many subsequent calls from this application to Windows.	*/

	hInst = hInstance;

    /* Open the datafile. OpenFile searches the path			*/
    hFile=OpenFile("COOKIE.S",&OfStruct,OF_READ);
    if (hFile==-1){
    	MessageBox(NULL,"No Cookiefile",
    		NULL,MB_OK|MB_ICONHAND);
    	return (FALSE);
    	}
    /* Allocate memory to hold cookie							*/
    hBuffer=LocalAlloc(LMEM_MOVEABLE,4096);
    if (!hBuffer){
    	MessageBox(NULL,"Not Enough Memory",
    		NULL,MB_OK|MB_ICONHAND);
    	return (FALSE);
    	}
    /* Lock the memory and obtain pointer						*/
    pBuffer=LocalLock(hBuffer);

    /* Prime the random-number generator with the system time	*/
    /* else we would get the same cookie every time. Note that	*/
    /* rand() is rather simple minded so you might get the same	*/
    /* cookie anyway if you run the program a couple of times.	*/
    /* in my DOS-cookie I used a Lattice-supplied 48-bit random	*/
    /* function. The random-number generator is primed with the	*/
    /* system time in seconds shifted four bits left. This will	*/
    /* give more varying seeds if the cookie is started several	*/
    /* times in succession. This reduces the likelyhood of the	*/
    /* user getting the same cookie twice in succession			*/

    temp=(unsigned) time(NULL);
    temp=temp<<4;
    srand(temp);

    /* Then calculate pointer into file and lseek()				*/

    fstat(hFile,&FileStatus);
    lFilePos=(long)rand();
    lFilePos=lFilePos*FileStatus.st_size;
    lFilePos=lFilePos/(LONG)RAND_MAX;
    lseek(hFile,lFilePos,SEEK_SET);

    iStrings=0;
    iBytes=_lread(hFile,pBuffer,4096);
    /* If ScanCookie() returns false then no cookie has been found	*/
    /* start again at top of file									*/
	if (!ScanCookie(iBytes)){
		lseek(hFile,0L,SEEK_SET);
        iStrings=0;
    	iBytes=_lread(hFile,pBuffer,4096);
    	ScanCookie(iBytes);
    	}
    _lclose(hFile);
	/* Create a main window for this application instance.	*/

	hWnd = CreateWindow(
		"CookieClass",			/* See RegisterClass() call */
		"Windows Cookie",		/* Text for window title bar */
		WS_OVERLAPPEDWINDOW,		/* Window style */
		CW_USEDEFAULT,			/* Default positions */
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		NULL,					/* Overlapped windows have no parent */
		NULL,					/* Use the window class menu */
		hInstance,				/* This instance owns this window */
		NULL					/* Pointer not needed */
    );

	/* If window could not be created, return "failure" */

	if (!hWnd)
		return (FALSE);

	/* Add a separator and an "About" entry to the system menu	*/
    hMenu=GetSystemMenu(hWnd,0);
    AppendMenu(hMenu,MF_SEPARATOR,0,NULL);
    AppendMenu(hMenu,MF_STRING,IDM_ABOUT,"About...");

	/* Make the window visible; update its client area; and return "success" */

	ShowWindow(hWnd, nCmdShow);		/* Show the window 			*/

	return (TRUE);	/* Returns the value from PostQuitMessage 	*/
}

/****************************************************************************

	FUNCTION: MainWndProc(HWND, unsigned, WORD, LONG)

	PURPOSE:  Processes messages

	MESSAGES:

		WM_COMMAND	  - application menu (OK Button)
		WM_SYSCOMMAND - System commands (About..)
		WM_DESTROY	  - destroy window
        WM_CREATE     - Create window. Create a child-window (OK-button)
        WM_PAINT      - Print the actual cookie
	COMMENTS:


****************************************************************************/

LONG FAR PASCAL MainWndProc(HWND hWnd, unsigned message, WORD wParam, LONG lParam)
{

	switch (message)
	{
		case WM_SYSCOMMAND:
			if (wParam==IDM_ABOUT){
				lpfnAbout=MakeProcInstance((FARPROC) About, hInst);
				DialogBox(hInst,"ABOUTBOX",hWnd,lpfnAbout);
				FreeProcInstance(lpfnAbout);
				}
			else
				return (DefWindowProc(hWnd, message, wParam, lParam));


		case WM_COMMAND:	/* message: command from application	*/
            if (wParam==IDOK){	/* User pressed OK, end application	*/
            		SendMessage(hWnd,WM_CLOSE,0,0L);
            		break;
            		}
            else
				return (DefWindowProc(hWnd, message, wParam, lParam));

        case WM_CHAR:
            SendMessage(hWnd,WM_CLOSE,0,0L);
            break;

		case WM_CREATE:
			/* Set up the timer to send us a message after a time 		*/
			/* determined by the number of lines in the cookie.			*/
			/* No need to check the function return since our app will 	*/
			/* run  OK without the timer.								*/

			idTimer = SetTimer(hWnd,1,5000*iStrings,(FARPROC) NULL);
			iButtonWidth=60;
			iButtonHeight=25;

			hButton=CreateWindow("button","OK",
					WS_CHILD|WS_VISIBLE|BS_PUSHBUTTON,
					10,10,
					iButtonWidth,iButtonHeight,
					hWnd,IDOK,
					hInst,NULL);
			PaintWindow(hWnd,TRUE);
			ShowWindow(hWnd,SW_SHOWNORMAL);
			UpdateWindow(hWnd);
			return (0);

        case WM_TIMER:			   /* The first timer-tick simply closes	*/
		case WM_DESTROY:		  /* message: window being destroyed */
			LocalUnlock(hBuffer); /* Unlock the memoty holding the cookie	*/
			LocalFree(hBuffer);   /* and free it							*/
			KillTimer(hWnd,idTimer);
			PostQuitMessage(0);
			break;

        case WM_PAINT:
        	ShowWindow(hWnd,SW_SHOW);
            PaintWindow(hWnd,FALSE);
            return (0);

    /* Message not processed. Call default handler	*/
        default:
        	return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	/* Message processed. Return NULL	*/
	return (NULL);
}
/****************************************************************************

	FUNCTION: PaintWindow(HWND,bReSize)

	PURPOSE:  Print the cookie-text

	COMMENTS: The size needed for the window is not known until after
	          it is printed, hence the resize


****************************************************************************/

void PaintWindow (HWND hWnd,BOOL bReSize)
			{

		int iWidth;
		int iMaxWidth;
		int nDrawX;
		int nDrawY;
		int iMargin;
		int i,ret;

		DWORD dwSize;
		char szText[300];
		HFONT hFont;
		HFONT hOldFont;
        TEXTMETRIC textmetric;

			hDC = BeginPaint (hWnd, &ps);

		/* Get the size characteristics of the current font	*/
		/* so that we can calculate vertical spacing		*/

			GetTextMetrics (hDC, &textmetric);

		/* Initialize drawing position to 1/4 inch from the top  */
		/* and from the left of the top, left corner of the      */
		/* client area of the main windows.                      */

			nDrawX = iMargin = GetDeviceCaps (hDC, LOGPIXELSX) / 4;
			nDrawY = GetDeviceCaps (hDC, LOGPIXELSY) / 4;
            iButtonHeight= 7 * (textmetric.tmHeight+textmetric.tmExternalLeading) / 4;
            iButtonWidth = textmetric.tmAveCharWidth * 10;

		/* Print the actual cookie-text. Note that COOKIE.S might	*/
		/* contain extended characters from the OEM-set, hence the	*/
		/* OemToAnsi call for each string. The maximum width is		*/
		/* calculated so that we can resize the window.				*/

			iMaxWidth=0;
            for (i=0;i<iStrings;i++){
            	OemToAnsi((LPSTR)(String[i]),(LPSTR)(szText));
				TextOut (hDC, nDrawX, nDrawY, szText, strlen (szText));
				dwSize=GetTextExtent (hDC, (LPSTR)szText, strlen (szText));
				iWidth=LOWORD(dwSize);
				if (iMaxWidth<iWidth) iMaxWidth=iWidth;
				nDrawY += textmetric.tmExternalLeading + textmetric.tmHeight;
                }

			EndPaint (hWnd,  &ps);

		/* Resize main and child windows	*/
			iMaxWidth+=2*iMargin;
			if (iMaxWidth<300) iMaxWidth=300;
			if (bReSize){
				MoveWindow(hWnd, 10,10,
					iMaxWidth,nDrawY+100,TRUE);

                MoveWindow(hButton,
					(iMaxWidth/2)-(iButtonWidth/2),nDrawY+40,
					iButtonWidth,iButtonHeight,TRUE);
					}
			SendMessage(hButton,WM_SETFOCUS,0,0L);
			}
/****************************************************************************

	FUNCTION: ScanCookie(iBytes)

	PURPOSE:  Scan the buffer looking for a Cookie


****************************************************************************/

BOOL ScanCookie(int iBytes){
int i;
BOOL bStarted;

    bStarted=FALSE;
    for(i=0;i<iBytes;i++){

    	switch(pBuffer[i]){
    		case 1:
    			if (bStarted){
   					pBuffer[i]=0;
    				i=iBytes;
    				}
    			else{
    			    String[iStrings]=pBuffer+2+i;
    			    bStarted=TRUE;
    			    }
    			break;
    		case 0x0D:
    			pBuffer[i]=0;
    			if (bStarted) String[++iStrings]=pBuffer+2+i;
    			if (iStrings>80) i=iBytes;
    			break;
    		case 0x0A:
    			pBuffer[i]=0;
    			break;
    			}
    		}
    	return (bStarted);
       }
/****************************************************************************

	FUNCTION: About(HWND, unsigned, WORD, LONG)

	PURPOSE:  Processes messages for "About" dialog box

	MESSAGES:

	WM_INITDIALOG - initialize dialog box
	WM_COMMAND    - Input received

****************************************************************************/

BOOL FAR PASCAL About(hDlg, message, wParam, lParam)
HWND hDlg;
unsigned message;
WORD wParam;
LONG lParam;
{
	switch (message) {
	case WM_INITDIALOG:
		return (TRUE);

	case WM_COMMAND:
		if (wParam == IDOK) {
		EndDialog(hDlg, TRUE);
		return (TRUE);
		}
		break;
	}
	return (FALSE);
}